from typing import Literal, Optional, Sequence, get_args, cast
import random
import secrets

from rapidata.api_client.models.ab_test_selection_a_inner import AbTestSelectionAInner
from rapidata.api_client.models.and_user_filter_model_filters_inner import (
    AndUserFilterModelFiltersInner,
)
from rapidata.api_client.models.create_order_model import CreateOrderModel
from rapidata.api_client.models.create_order_model_referee import (
    CreateOrderModelReferee,
)
from rapidata.api_client.models.create_order_model_workflow import (
    CreateOrderModelWorkflow,
)
from rapidata.api_client.models.sticky_state import StickyState

from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.rapidata_client.exceptions.failed_upload_exception import (
    FailedUploadException,
)
from rapidata.rapidata_client.filter import RapidataFilter, UserScoreFilter
from rapidata.rapidata_client.config import (
    logger,
    managed_print,
    rapidata_config,
    tracer,
)
from rapidata.rapidata_client.validation.validation_set_manager import (
    ValidationSetManager,
)
from rapidata.rapidata_client.validation.rapidata_validation_set import (
    RapidataValidationSet,
)
from rapidata.rapidata_client.order.dataset._rapidata_dataset import RapidataDataset
from rapidata.rapidata_client.order.rapidata_order import RapidataOrder
from rapidata.rapidata_client.referee import Referee
from rapidata.rapidata_client.referee._naive_referee import NaiveReferee
from rapidata.rapidata_client.selection._base_selection import RapidataSelection
from rapidata.rapidata_client.settings import RapidataSetting
from rapidata.rapidata_client.workflow import (
    Workflow,
    CompareWorkflow,
    ClassifyWorkflow,
)
from rapidata.rapidata_client.selection import (
    ConditionalValidationSelection,
    LabelingSelection,
    CappedSelection,
)
from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.api.rapidata_api_client import (
    suppress_rapidata_error_logging,
)

StickyStateLiteral = Literal["Temporary", "Permanent", "Passive"]


class RapidataOrderBuilder:
    """Builder object for creating Rapidata orders.

    Use the fluent interface to set the desired configuration. Add a workflow to the order using `.workflow()` and finally call `.create()` to create the order.

    Args:
        openapi_service (OpenAPIService): The OpenAPIService instance.
        name (str): The name of the order.
    """

    def __init__(
        self,
        name: str,
        openapi_service: OpenAPIService,
    ):
        self._name = name
        self._openapi_service = openapi_service
        self.__dataset: Optional[RapidataDataset] = None
        self.__workflow: Workflow | None = None
        self.__referee: Referee | None = None
        self.__validation_set: RapidataValidationSet | None = None
        self.__settings: Sequence[RapidataSetting] | None = None
        self.__user_filters: list[RapidataFilter] = []
        self.__selections: Sequence[RapidataSelection] = []
        self.__priority: int | None = None
        self.__datapoints: list[Datapoint] = []
        self.__sticky_state_value: StickyStateLiteral | None = None
        self.__temporary_sticky_enabled: bool = False
        self.__validation_set_manager: ValidationSetManager = ValidationSetManager(
            self._openapi_service
        )

    def _to_model(self) -> CreateOrderModel:
        """
        Convert the builder configuration to a CreateOrderModel.

        Raises:
            ValueError: If no workflow is provided.

        Returns:
            CreateOrderModel: The model representing the order configuration.
        """
        if self.__workflow is None:
            raise ValueError("You must provide a workflow to create an order.")

        if self.__referee is None:
            managed_print("No referee provided, using default NaiveReferee.")
            self.__referee = NaiveReferee()

        sticky_state = self.__sticky_state_value
        if not sticky_state and self.__temporary_sticky_enabled:
            sticky_state = "Temporary"
            logger.debug(
                "Setting sticky state to Temporary due to temporary sticky enabled."
            )

        validation_set_id = (
            self.__validation_set.id
            if (self.__validation_set and not self.__selections)
            else None
        )

        return CreateOrderModel(
            _t="CreateOrderModel",
            orderName=self._name,
            workflow=CreateOrderModelWorkflow(self.__workflow._to_model()),
            userFilters=[
                AndUserFilterModelFiltersInner(user_filter._to_model())
                for user_filter in self.__user_filters
            ],
            referee=CreateOrderModelReferee(self.__referee._to_model()),
            validationSetId=validation_set_id,
            featureFlags=(
                [setting._to_feature_flag() for setting in self.__settings]
                if self.__settings is not None
                else None
            ),
            selections=(
                [
                    AbTestSelectionAInner(selection._to_model())
                    for selection in self.__selections
                ]
                if self.__selections
                else None
            ),
            priority=self.__priority,
            stickyState=(StickyState(sticky_state) if sticky_state else None),
        )

    def _generate_id(self, length=9):
        """Optimal balance: fast comparisons + low collision risk"""
        alphabet = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
        return "".join(secrets.choice(alphabet) for _ in range(length))

    def _set_validation_set_id(self) -> None:
        """
        Get the validation set ID for the order.
        """
        assert self.__workflow is not None

        required_amount = min(int(len(self.__datapoints) * 0.01) or 3, 15)

        if self.__validation_set is None:
            try:
                with suppress_rapidata_error_logging():
                    val_set_id = (
                        (
                            self._openapi_service.validation_api.validation_set_recommended_get(
                                asset_type=[self.__datapoints[0].get_asset_type()],
                                modality=[self.__workflow.modality],
                                instruction=self.__workflow._get_instruction(),
                                prompt_type=[
                                    t.value
                                    for t in self.__datapoints[0].get_prompt_type()
                                ],
                            )
                        )
                        .validation_sets[0]
                        .id
                    )
                    self.__validation_set = (
                        self.__validation_set_manager.get_validation_set_by_id(
                            val_set_id
                        )
                    )

            except Exception as e:
                logger.debug("No recommended validation set found, error: %s", e)

        sufficient_rapids_count = False
        if self.__validation_set is not None:
            sufficient_rapids_count = (
                self.__validation_set_manager._get_total_and_labeled_rapids_count(
                    self.__validation_set.id
                )[1]
                >= required_amount
            )

        if self.__validation_set is None or not sufficient_rapids_count:
            if (
                len(self.__datapoints)
                < rapidata_config.order.minOrderDatapointsForValidation
            ):
                logger.debug(
                    "No recommended validation set found, dataset too small to create one."
                )
                return

            logger.info("No recommended validation set found, creating new one.")
            managed_print()
            managed_print(
                f"No recommended validation set found, new one will be created."
            )

            new_dimension = self._generate_id()
            logger.debug("New dimension created: %s", new_dimension)
            rng = random.Random(42)
            self.__validation_set = (
                self.__validation_set_manager._create_order_validation_set(
                    workflow=self.__workflow,
                    name=self._name,
                    datapoints=rng.sample(
                        self.__datapoints, len(self.__datapoints)
                    ),  # shuffle the datapoints with a specific seed
                    required_amount=required_amount,
                    settings=self.__settings,
                    dimensions=[new_dimension],
                )
            )

        self.__validation_set.update_should_alert(False)
        self.__validation_set.update_can_be_flagged(False)

        logger.debug("New validation set created for order: %s", self.__validation_set)

        self.__selections = [
            CappedSelection(
                selections=[
                    ConditionalValidationSelection(
                        validation_set_id=self.__validation_set.id,
                        dimensions=self.__validation_set.dimensions,
                        thresholds=[0, 0.5, 0.7],
                        chances=[1, 1, 0.2],
                        rapid_counts=[10, 1, 1],
                    ),
                    LabelingSelection(amount=1),
                ],
                max_rapids=3,
            )
        ]

        for dimension in self.__validation_set.dimensions:
            self.__user_filters.append(
                UserScoreFilter(
                    lower_bound=0.3,
                    upper_bound=1,
                    dimension=dimension,
                )
            )

    def _create(self) -> RapidataOrder:
        """
        Create the Rapidata order by making the necessary API calls based on the builder's configuration.

        Raises:
            ValueError: If both media paths and texts are provided, or if neither is provided.
            AssertionError: If the workflow is a CompareWorkflow and media paths are not in pairs.

        Returns:
            RapidataOrder: The created RapidataOrder instance.
        """
        if (
            rapidata_config.order.autoValidationSetCreation
            and isinstance(self.__workflow, (CompareWorkflow, ClassifyWorkflow))
            and not self.__selections
            and rapidata_config.enableBetaFeatures
        ):
            self._set_validation_set_id()
            self.__temporary_sticky_enabled = True
            logger.debug("Temporary sticky enabled for order creation.")

        order_model = self._to_model()
        logger.debug("Creating order with model: %s", order_model)

        self.__temporary_sticky_enabled = False
        logger.debug("Disabling temporary sticky after order creation.")

        result = self._openapi_service.order_api.order_post(
            create_order_model=order_model
        )

        order = RapidataOrder(
            order_id=str(result.order_id),
            openapi_service=self._openapi_service,
            name=self._name,
        )
        logger.debug("Order created: %s", order)

        if not result.dataset_id:
            logger.error("No Dataset was created for order %s", order.id)
            return order

        self.__dataset = RapidataDataset(result.dataset_id, self._openapi_service)

        logger.debug("Dataset created with ID: %s", self.__dataset.id)
        logger.debug("Adding datapoints to the order.")
        with tracer.start_as_current_span("add_datapoints"):
            _, failed_uploads = self.__dataset.add_datapoints(self.__datapoints)

            if failed_uploads:
                raise FailedUploadException(self.__dataset, order, failed_uploads)

        logger.debug("Datapoints added to the order.")
        logger.debug("Setting order to preview")
        try:
            self._openapi_service.order_api.order_order_id_preview_post(order.id)
        except Exception:
            raise FailedUploadException(self.__dataset, order, failed_uploads)
        return order

    def _workflow(self, workflow: Workflow) -> "RapidataOrderBuilder":
        """
        Set the workflow for the order.

        Args:
            workflow (Workflow): The workflow to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(workflow, Workflow):
            raise TypeError("Workflow must be of type Workflow.")

        self.__workflow = workflow
        return self

    def _referee(self, referee: Referee) -> "RapidataOrderBuilder":
        """
        Set the referee for the order.

        Args:
            referee (Referee): The referee to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(referee, Referee):
            raise TypeError("Referee must be of type Referee.")

        self.__referee = referee
        return self

    def _datapoints(
        self,
        datapoints: list[Datapoint],
    ) -> "RapidataOrderBuilder":
        """
        Set the datapoints for the order.

        Args:
            datapoints: (Sequence[Datapoint]): The datapoints to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(datapoints, list):
            raise TypeError(
                "Datapoints must be provided as a list of Datapoint objects."
            )

        self.__datapoints = datapoints
        return self

    def _settings(self, settings: Sequence[RapidataSetting]) -> "RapidataOrderBuilder":
        """
        Set the settings for the order.

        Args:
            settings (Settings): The settings to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """

        if not isinstance(settings, list):
            raise TypeError("Settings must be provided as a list of Setting objects.")

        for s in settings:
            if not isinstance(s, RapidataSetting):
                raise TypeError("The settings list must only contain Setting objects.")

        self.__settings = settings
        return self

    def _filters(self, filters: Sequence[RapidataFilter]) -> "RapidataOrderBuilder":
        """
        Set the filters for the order, e.g., country, language, userscore, etc.

        Args:
            filters (Sequence[Filters]): The user filters to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(filters, list):
            raise TypeError("Filters must be provided as a list of Filter objects.")

        for f in filters:
            if not isinstance(f, RapidataFilter):
                raise TypeError("Filters must be of type Filter.")

        if len(self.__user_filters) > 0:
            managed_print("Overwriting existing user filters.")

        self.__user_filters = filters
        return self

    def _validation_set_id(
        self, validation_set_id: str | None = None
    ) -> "RapidataOrderBuilder":
        """
        Set the validation set ID for the order.

        Args:
            validation_set_id (str): The validation set ID to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if validation_set_id is not None and not isinstance(validation_set_id, str):
            raise TypeError("Validation set ID must be of type str.")
        if validation_set_id is not None:
            self.__validation_set = (
                self.__validation_set_manager.get_validation_set_by_id(
                    validation_set_id
                )
            )
            return self
        self.__validation_set = validation_set_id
        return self

    def _rapids_per_bag(self, amount: int) -> "RapidataOrderBuilder":
        """
        Define the number of tasks a user sees in a single session.

        Args:
            amount (int): The number of tasks a user sees in a single session.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.

        Raises:
            NotImplementedError: This method is not implemented yet.
        """
        raise NotImplementedError("Not implemented yet.")

    def _selections(
        self, selections: Sequence[RapidataSelection]
    ) -> "RapidataOrderBuilder":
        """
        Set the selections for the order.

        Args:
            selections (Sequence[Selection]): The selections to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if not isinstance(selections, list):
            raise TypeError(
                "Selections must be provided as a list of Selection objects."
            )

        for selection in selections:
            if not isinstance(selection, RapidataSelection):
                raise TypeError("Selections must be of type Selection.")

        self.__selections = selections  # type: ignore
        return self

    def _priority(self, priority: int | None = None) -> "RapidataOrderBuilder":
        """
        Set the priority for the order.

        Args:
            priority (int): The priority to be set.

        Returns:
            RapidataOrderBuilder: The updated RapidataOrderBuilder instance.
        """
        if priority is not None and not isinstance(priority, int):
            raise TypeError("Priority must be of type int.")

        self.__priority = priority
        return self

    def _sticky_state(
        self, sticky_state: StickyStateLiteral | None = None
    ) -> "RapidataOrderBuilder":
        """
        Set the sticky state for the order.
        """
        sticky_state_valid_values = get_args(StickyStateLiteral)

        if sticky_state is not None and sticky_state not in sticky_state_valid_values:
            raise ValueError(
                f"Sticky state must be one of {sticky_state_valid_values} or None"
            )

        self.__sticky_state_value = sticky_state
        return self



# coding: utf-8

"""
    Rapidata.Dataset

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: v1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
import pprint
from pydantic import BaseModel, ConfigDict, Field, StrictStr, ValidationError, field_validator
from typing import Any, List, Optional
from rapidata.api_client.models.campaign_filter import CampaignFilter
from rapidata.api_client.models.country_filter import CountryFilter
from rapidata.api_client.models.demographic_filter import DemographicFilter
from rapidata.api_client.models.language_filter import LanguageFilter
from rapidata.api_client.models.new_user_filter import NewUserFilter
from rapidata.api_client.models.response_count_filter import ResponseCountFilter
from rapidata.api_client.models.user_action_restriction_filter import UserActionRestrictionFilter
from rapidata.api_client.models.user_score_filter import UserScoreFilter
from rapidata.api_client.models.user_state_filter import UserStateFilter
from pydantic import StrictStr, Field
from typing import Union, List, Set, Optional, Dict
from typing_extensions import Literal, Self

ANDFILTERFILTERSINNER_ONE_OF_SCHEMAS = ["AndFilter", "CampaignFilter", "CountryFilter", "DemographicFilter", "LanguageFilter", "NewUserFilter", "NotFilter", "OrFilter", "ResponseCountFilter", "UserActionRestrictionFilter", "UserScoreFilter", "UserStateFilter"]

class AndFilterFiltersInner(BaseModel):
    """
    AndFilterFiltersInner
    """
    # data type: AndFilter
    oneof_schema_1_validator: Optional[AndFilter] = None
    # data type: CampaignFilter
    oneof_schema_2_validator: Optional[CampaignFilter] = None
    # data type: CountryFilter
    oneof_schema_3_validator: Optional[CountryFilter] = None
    # data type: DemographicFilter
    oneof_schema_4_validator: Optional[DemographicFilter] = None
    # data type: LanguageFilter
    oneof_schema_5_validator: Optional[LanguageFilter] = None
    # data type: NewUserFilter
    oneof_schema_6_validator: Optional[NewUserFilter] = None
    # data type: NotFilter
    oneof_schema_7_validator: Optional[NotFilter] = None
    # data type: OrFilter
    oneof_schema_8_validator: Optional[OrFilter] = None
    # data type: ResponseCountFilter
    oneof_schema_9_validator: Optional[ResponseCountFilter] = None
    # data type: UserActionRestrictionFilter
    oneof_schema_10_validator: Optional[UserActionRestrictionFilter] = None
    # data type: UserScoreFilter
    oneof_schema_11_validator: Optional[UserScoreFilter] = None
    # data type: UserStateFilter
    oneof_schema_12_validator: Optional[UserStateFilter] = None
    actual_instance: Optional[Union[AndFilter, CampaignFilter, CountryFilter, DemographicFilter, LanguageFilter, NewUserFilter, NotFilter, OrFilter, ResponseCountFilter, UserActionRestrictionFilter, UserScoreFilter, UserStateFilter]] = None
    one_of_schemas: Set[str] = { "AndFilter", "CampaignFilter", "CountryFilter", "DemographicFilter", "LanguageFilter", "NewUserFilter", "NotFilter", "OrFilter", "ResponseCountFilter", "UserActionRestrictionFilter", "UserScoreFilter", "UserStateFilter" }

    model_config = ConfigDict(
        validate_assignment=True,
        protected_namespaces=(),
    )


    discriminator_value_class_map: Dict[str, str] = {
    }

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError("If a position argument is used, only 1 is allowed to set `actual_instance`")
            if kwargs:
                raise ValueError("If a position argument is used, keyword arguments cannot be used.")
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator('actual_instance')
    def actual_instance_must_validate_oneof(cls, v):
        instance = AndFilterFiltersInner.model_construct()
        error_messages = []
        match = 0
        # validate data type: AndFilter
        if not isinstance(v, AndFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `AndFilter`")
        else:
            match += 1
        # validate data type: CampaignFilter
        if not isinstance(v, CampaignFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `CampaignFilter`")
        else:
            match += 1
        # validate data type: CountryFilter
        if not isinstance(v, CountryFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `CountryFilter`")
        else:
            match += 1
        # validate data type: DemographicFilter
        if not isinstance(v, DemographicFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `DemographicFilter`")
        else:
            match += 1
        # validate data type: LanguageFilter
        if not isinstance(v, LanguageFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `LanguageFilter`")
        else:
            match += 1
        # validate data type: NewUserFilter
        if not isinstance(v, NewUserFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `NewUserFilter`")
        else:
            match += 1
        # validate data type: NotFilter
        if not isinstance(v, NotFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `NotFilter`")
        else:
            match += 1
        # validate data type: OrFilter
        if not isinstance(v, OrFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `OrFilter`")
        else:
            match += 1
        # validate data type: ResponseCountFilter
        if not isinstance(v, ResponseCountFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `ResponseCountFilter`")
        else:
            match += 1
        # validate data type: UserActionRestrictionFilter
        if not isinstance(v, UserActionRestrictionFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `UserActionRestrictionFilter`")
        else:
            match += 1
        # validate data type: UserScoreFilter
        if not isinstance(v, UserScoreFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `UserScoreFilter`")
        else:
            match += 1
        # validate data type: UserStateFilter
        if not isinstance(v, UserStateFilter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `UserStateFilter`")
        else:
            match += 1
        if match > 1:
            # more than 1 match
            raise ValueError("Multiple matches found when setting `actual_instance` in AndFilterFiltersInner with oneOf schemas: AndFilter, CampaignFilter, CountryFilter, DemographicFilter, LanguageFilter, NewUserFilter, NotFilter, OrFilter, ResponseCountFilter, UserActionRestrictionFilter, UserScoreFilter, UserStateFilter. Details: " + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError("No match found when setting `actual_instance` in AndFilterFiltersInner with oneOf schemas: AndFilter, CampaignFilter, CountryFilter, DemographicFilter, LanguageFilter, NewUserFilter, NotFilter, OrFilter, ResponseCountFilter, UserActionRestrictionFilter, UserScoreFilter, UserStateFilter. Details: " + ", ".join(error_messages))
        else:
            return v

    @classmethod
    def from_dict(cls, obj: Union[str, Dict[str, Any]]) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        match = 0

        # deserialize data into AndFilter
        try:
            instance.actual_instance = AndFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into CampaignFilter
        try:
            instance.actual_instance = CampaignFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into CountryFilter
        try:
            instance.actual_instance = CountryFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into DemographicFilter
        try:
            instance.actual_instance = DemographicFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into LanguageFilter
        try:
            instance.actual_instance = LanguageFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into NewUserFilter
        try:
            instance.actual_instance = NewUserFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into NotFilter
        try:
            instance.actual_instance = NotFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into OrFilter
        try:
            instance.actual_instance = OrFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into ResponseCountFilter
        try:
            instance.actual_instance = ResponseCountFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into UserActionRestrictionFilter
        try:
            instance.actual_instance = UserActionRestrictionFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into UserScoreFilter
        try:
            instance.actual_instance = UserScoreFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into UserStateFilter
        try:
            instance.actual_instance = UserStateFilter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if match > 1:
            # more than 1 match
            raise ValueError("Multiple matches found when deserializing the JSON string into AndFilterFiltersInner with oneOf schemas: AndFilter, CampaignFilter, CountryFilter, DemographicFilter, LanguageFilter, NewUserFilter, NotFilter, OrFilter, ResponseCountFilter, UserActionRestrictionFilter, UserScoreFilter, UserStateFilter. Details: " + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError("No match found when deserializing the JSON string into AndFilterFiltersInner with oneOf schemas: AndFilter, CampaignFilter, CountryFilter, DemographicFilter, LanguageFilter, NewUserFilter, NotFilter, OrFilter, ResponseCountFilter, UserActionRestrictionFilter, UserScoreFilter, UserStateFilter. Details: " + ", ".join(error_messages))
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(self.actual_instance.to_json):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(self) -> Optional[Union[Dict[str, Any], AndFilter, CampaignFilter, CountryFilter, DemographicFilter, LanguageFilter, NewUserFilter, NotFilter, OrFilter, ResponseCountFilter, UserActionRestrictionFilter, UserScoreFilter, UserStateFilter]]:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(self.actual_instance.to_dict):
            return self.actual_instance.to_dict()
        else:
            # primitive type
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())

from rapidata.api_client.models.and_filter import AndFilter
from rapidata.api_client.models.not_filter import NotFilter
from rapidata.api_client.models.or_filter import OrFilter
# TODO: Rewrite to not use raise_errors
AndFilterFiltersInner.model_rebuild(raise_errors=False)




from typing import Any
from rapidata.api_client.models.simple_workflow_model import SimpleWorkflowModel
from rapidata.api_client.models.simple_workflow_model_blueprint import (
    SimpleWorkflowModelBlueprint,
)
from rapidata.rapidata_client.workflow import Workflow
from rapidata.api_client.models.free_text_rapid_blueprint import FreeTextRapidBlueprint
from rapidata.api_client import FreeTextPayload
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.api_client.models.rapid_modality import RapidModality


class FreeTextWorkflow(Workflow):
    """
    A workflow for free text input tasks.

    This class represents a workflow where users can provide free-form text responses
    to a given instruction.

    Attributes:
        _instruction (str): The instruction to be answered with free text.

    Args:
        instruction (str): The instruction to be presented for free text input.
        validation_system_prompt (str): The system prompt to determine if the provided free text response is spam or not.
            Should always specify that the LLM should respond with 'not spam' or 'spam'.
    """

    modality = RapidModality.FREETEXT

    def __init__(self, instruction: str, validation_system_prompt: str | None = None):
        super().__init__(type="SimpleWorkflowConfig")
        self._instruction = instruction
        self._validation_system_prompt = validation_system_prompt

    def _get_instruction(self) -> str:
        return self._instruction

    def _to_dict(self) -> dict[str, Any]:
        return {
            **super()._to_dict(),
            "blueprint": {
                "_t": "FreeTextBlueprint",
                "question": self._instruction,
                "validationSystemPrompt": self._validation_system_prompt,
            },
        }

    def _to_model(self) -> SimpleWorkflowModel:
        blueprint = FreeTextRapidBlueprint(
            _t="FreeTextBlueprint",
            question=self._instruction,
            validationSystemPrompt=self._validation_system_prompt,
        )

        return SimpleWorkflowModel(
            _t="SimpleWorkflow",
            blueprint=SimpleWorkflowModelBlueprint(blueprint),
        )

    def _to_payload(self, datapoint: Datapoint) -> FreeTextPayload:
        return FreeTextPayload(
            _t="FreeTextPayload",
            question=self._instruction,
        )

    def __str__(self) -> str:
        return f"FreeTextWorkflow(instruction='{self._instruction}', validation_system_prompt={self._validation_system_prompt})"

    def __repr__(self) -> str:
        return f"FreeTextWorkflow(instruction={self._instruction!r}, validation_system_prompt={self._validation_system_prompt!r})"



from rapidata.api_client.models.simple_workflow_model import SimpleWorkflowModel
from rapidata.api_client.models.simple_workflow_model_blueprint import (
    SimpleWorkflowModelBlueprint,
)
from rapidata.api_client.models.locate_rapid_blueprint import LocateRapidBlueprint
from rapidata.rapidata_client.workflow._base_workflow import Workflow
from rapidata.api_client import LocatePayload
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.api_client.models.rapid_modality import RapidModality


class LocateWorkflow(Workflow):
    modality = RapidModality.LOCATE

    def __init__(self, target: str):
        super().__init__(type="SimpleWorkflowConfig")
        self._target = target

    def _get_instruction(self) -> str:
        return self._target

    def _to_model(self) -> SimpleWorkflowModel:
        blueprint = LocateRapidBlueprint(_t="LocateBlueprint", target=self._target)

        return SimpleWorkflowModel(
            _t="SimpleWorkflow",
            blueprint=SimpleWorkflowModelBlueprint(blueprint),
        )

    def _to_payload(self, datapoint: Datapoint) -> LocatePayload:
        return LocatePayload(
            _t="LocatePayload",
            target=self._target,
        )

    def __str__(self) -> str:
        return f"LocateWorkflow(target='{self._target}')"

    def __repr__(self) -> str:
        return f"LocateWorkflow(target={self._target!r})"



from rapidata.api_client.models.simple_workflow_model import SimpleWorkflowModel
from rapidata.api_client.models.simple_workflow_model_blueprint import (
    SimpleWorkflowModelBlueprint,
)
from rapidata.api_client.models.validation_set_zip_post_request_blueprint import (
    ValidationSetZipPostRequestBlueprint,
)
from rapidata.api_client.models.line_rapid_blueprint import LineRapidBlueprint
from rapidata.rapidata_client.workflow._base_workflow import Workflow
from rapidata.api_client import LinePayload
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.api_client.models.rapid_modality import RapidModality


class DrawWorkflow(Workflow):
    modality = RapidModality.LINE

    def __init__(self, target: str):
        super().__init__(type="SimpleWorkflowConfig")
        self._target = target

    def _get_instruction(self) -> str:
        return self._target

    def _to_model(self) -> SimpleWorkflowModel:
        blueprint = LineRapidBlueprint(_t="LineBlueprint", target=self._target)

        return SimpleWorkflowModel(
            _t="SimpleWorkflow",
            blueprint=SimpleWorkflowModelBlueprint(blueprint),
        )

    def _to_payload(self, datapoint: Datapoint) -> LinePayload:
        return LinePayload(
            _t="LinePayload",
            target=self._target,
        )

    def __str__(self) -> str:
        return f"DrawWorkflow(target='{self._target}')"

    def __repr__(self) -> str:
        return f"DrawWorkflow(target={self._target!r})"



from rapidata.rapidata_client.settings._rapidata_setting import RapidataSetting


class CustomSetting(RapidataSetting):
    """
    Set a custom setting with the given key and value. Use this to enable features that do not have a dedicated method (yet)

    Args:
        key (str): The key for the custom setting.
        value (str): The value for the custom setting.
    """

    def __init__(self, key: str, value: str):
        if not isinstance(key, str):
            raise ValueError("The key must be a string.")

        super().__init__(key=key, value=value)



from .rapidata_client import RapidataClient
from .selection import (
    DemographicSelection,
    LabelingSelection,
    ValidationSelection,
    ConditionalValidationSelection,
    CappedSelection,
    ShufflingSelection,
    RapidataRetrievalMode,
    EffortSelection,
)
from .datapoints import Datapoint
from .datapoints.metadata import (
    PrivateTextMetadata,
    PublicTextMetadata,
    PromptMetadata,
    SelectWordsMetadata,
)
from .settings import (
    RapidataSettings,
    TranslationBehaviourOptions,
    AlertOnFastResponse,
    TranslationBehaviour,
    FreeTextMinimumCharacters,
    NoShuffle,
    PlayVideoUntilTheEnd,
    CustomSetting,
    AllowNeitherBoth,
    SwapContextInstruction,
    MuteVideo,
)
from .filter import (
    CountryFilter,
    LanguageFilter,
    UserScoreFilter,
    CampaignFilter,
    AgeFilter,
    GenderFilter,
    CustomFilter,
    AgeGroup,
    Gender,
    NotFilter,
    OrFilter,
    ResponseCountFilter,
)
from .validation import Box
from .exceptions import FailedUploadException
from .config import rapidata_config, logger, managed_print



from typing import Literal, Optional, Sequence, get_args

from rapidata.rapidata_client.config import logger
from rapidata.rapidata_client.config.tracer import tracer
from rapidata.rapidata_client.datapoints._asset_uploader import AssetUploader
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.rapidata_client.datapoints._datapoints_validator import (
    DatapointsValidator,
)
from rapidata.rapidata_client.filter import RapidataFilter
from rapidata.rapidata_client.filter.rapidata_filters import RapidataFilters
from rapidata.rapidata_client.order._rapidata_order_builder import (
    RapidataOrderBuilder,
    StickyStateLiteral,
)
from rapidata.rapidata_client.order.rapidata_order import RapidataOrder
from rapidata.rapidata_client.referee._early_stopping_referee import (
    EarlyStoppingReferee,
)
from rapidata.rapidata_client.referee._naive_referee import NaiveReferee
from rapidata.rapidata_client.selection._base_selection import RapidataSelection
from rapidata.rapidata_client.selection.rapidata_selections import RapidataSelections
from rapidata.rapidata_client.settings import RapidataSetting, RapidataSettings
from rapidata.rapidata_client.workflow import (
    ClassifyWorkflow,
    CompareWorkflow,
    DrawWorkflow,
    FreeTextWorkflow,
    LocateWorkflow,
    MultiRankingWorkflow,
    SelectWordsWorkflow,
    TimestampWorkflow,
    Workflow,
)
from rapidata.api_client.models.existing_asset_input import ExistingAssetInput
from rapidata.api_client.models.filter import Filter
from rapidata.api_client.models.filter_operator import FilterOperator
from rapidata.api_client.models.multi_asset_input_assets_inner import (
    MultiAssetInputAssetsInner,
)
from rapidata.api_client.models.page_info import PageInfo
from rapidata.api_client.models.query_model import QueryModel
from rapidata.api_client.models.root_filter import RootFilter
from rapidata.api_client.models.sort_criterion import SortCriterion
from rapidata.api_client.models.sort_direction import SortDirection
from rapidata.service.openapi_service import OpenAPIService


class RapidataOrderManager:
    """
    Handels everything regarding the orders from creation to retrieval.

    Attributes:
        filters (RapidataFilters): The RapidataFilters instance.
        settings (RapidataSettings): The RapidataSettings instance.
        selections (RapidataSelections): The RapidataSelections instance."""

    def __init__(self, openapi_service: OpenAPIService):
        self.__openapi_service = openapi_service
        self.filters = RapidataFilters
        self.settings = RapidataSettings
        self.selections = RapidataSelections
        self.__priority: int | None = None
        self.__sticky_state: StickyStateLiteral | None = None
        self._asset_uploader = AssetUploader(openapi_service)
        logger.debug("RapidataOrderManager initialized")

    def _create_general_order(
        self,
        name: str,
        workflow: Workflow,
        datapoints: list[Datapoint],
        responses_per_datapoint: int = 10,
        validation_set_id: str | None = None,
        confidence_threshold: float | None = None,
        filters: Sequence[RapidataFilter] | None = None,
        settings: Sequence[RapidataSetting] | None = None,
        selections: Sequence[RapidataSelection] | None = None,
    ) -> RapidataOrder:
        if filters is None:
            filters = []
        if settings is None:
            settings = []
        if selections is None:
            selections = []

        if not confidence_threshold:
            referee = NaiveReferee(responses=responses_per_datapoint)
        else:
            referee = EarlyStoppingReferee(
                threshold=confidence_threshold,
                max_vote_count=responses_per_datapoint,
            )

        logger.debug(
            "Creating order with parameters: name %s, workflow %s, datapoints %s, responses_per_datapoint %s, validation_set_id %s, confidence_threshold %s, filters %s, settings %s, selections %s",
            name,
            workflow,
            datapoints,
            responses_per_datapoint,
            validation_set_id,
            confidence_threshold,
            filters,
            settings,
            selections,
        )

        order_builder = RapidataOrderBuilder(
            name=name, openapi_service=self.__openapi_service
        )

        if selections and validation_set_id:
            logger.warning(
                "Warning: Both selections and validation_set_id provided. Ignoring validation_set_id."
            )

        order = (
            order_builder._workflow(workflow)
            ._datapoints(datapoints=datapoints)
            ._referee(referee)
            ._filters(filters)
            ._selections(selections)
            ._settings(settings)
            ._validation_set_id(validation_set_id if not selections else None)
            ._priority(self.__priority)
            ._sticky_state(self.__sticky_state)
            ._create()
        )
        logger.debug("Order created: %s", order)
        return order

    def _set_priority(self, priority: int | None):
        if priority is not None and not isinstance(priority, int):
            raise TypeError("Priority must be an integer or None")

        if priority is not None and priority < 0:
            raise ValueError("Priority must be greater than 0 or None")

        self.__priority = priority

    def _set_sticky_state(self, sticky_state: StickyStateLiteral | None):
        sticky_state_valid_values = get_args(StickyStateLiteral)

        if sticky_state is not None and sticky_state not in sticky_state_valid_values:
            raise ValueError(
                f"Sticky state must be one of {sticky_state_valid_values} or None"
            )

        self.__sticky_state = sticky_state

    def create_classification_order(
        self,
        name: str,
        instruction: str,
        answer_options: list[str],
        datapoints: list[str],
        data_type: Literal["media", "text"] = "media",
        responses_per_datapoint: int = 10,
        contexts: list[str] | None = None,
        media_contexts: list[str] | None = None,
        validation_set_id: str | None = None,
        confidence_threshold: float | None = None,
        filters: Sequence[RapidataFilter] | None = None,
        settings: Sequence[RapidataSetting] | None = None,
        selections: Sequence[RapidataSelection] | None = None,
        private_notes: list[str] | None = None,
    ) -> RapidataOrder:
        """Create a classification order.

        With this order you can have a datapoint (image, text, video, audio) be classified into one of the answer options.
        Each response will be exactly one of the answer options.

        Args:
            name (str): The name of the order. (Will not be shown to the labeler)
            instruction (str): The instruction for how the data should be classified.
            answer_options (list[str]): The list of options for the classification.
            datapoints (list[str]): The list of datapoints for the classification - each datapoint will be labeled.
            data_type (str, optional): The data type of the datapoints. Defaults to "media" (any form of image, video or audio). \n
                Other option: "text".
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the classification. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction and options. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            media_contexts (list[str], optional): The list of media contexts for the classification i.e links to the images / videos. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction and options. (Therefore will be different for each datapoint)
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            confidence_threshold (float, optional): The probability threshold for the classification. Defaults to None.\n
                If provided, the classification datapoint will stop after the threshold is reached or at the number of responses, whatever happens first.
            filters (Sequence[RapidataFilter], optional): The list of filters for the classification. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the classification. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the classification. Defaults to []. Decides in what order the tasks should be shown.
            private_notes (list[str], optional): The list of private notes for the classification. Defaults to None.
                If provided has to be the same length as datapoints.\n
                This will NOT be shown to the labelers but will be included in the result purely for your own reference.
        """
        with tracer.start_as_current_span(
            "RapidataOrderManager.create_classification_order"
        ):
            if not isinstance(datapoints, list) or not all(
                isinstance(datapoint, str) for datapoint in datapoints
            ):
                raise ValueError("Datapoints must be a list of strings")

            datapoints_instances = DatapointsValidator.map_datapoints(
                datapoints=datapoints,
                contexts=contexts,
                media_contexts=media_contexts,
                private_notes=private_notes,
                data_type=data_type,
            )
            return self._create_general_order(
                name=name,
                workflow=ClassifyWorkflow(
                    instruction=instruction, answer_options=answer_options
                ),
                datapoints=datapoints_instances,
                responses_per_datapoint=responses_per_datapoint,
                validation_set_id=validation_set_id,
                confidence_threshold=confidence_threshold,
                filters=filters,
                selections=selections,
                settings=settings,
            )

    def create_compare_order(
        self,
        name: str,
        instruction: str,
        datapoints: list[list[str]],
        data_type: Literal["media", "text"] = "media",
        responses_per_datapoint: int = 10,
        contexts: list[str] | None = None,
        media_contexts: list[str] | None = None,
        a_b_names: list[str] | None = None,
        validation_set_id: str | None = None,
        confidence_threshold: float | None = None,
        filters: Sequence[RapidataFilter] | None = None,
        settings: Sequence[RapidataSetting] | None = None,
        selections: Sequence[RapidataSelection] | None = None,
        private_notes: list[str] | None = None,
    ) -> RapidataOrder:
        """Create a compare order.

        With this order you compare two datapoints (image, text, video, audio) and the annotators will choose one of the two based on the instruction.

        Args:
            name (str): The name of the order. (Will not be shown to the labeler)
            instruction (str): The instruction for the comparison. Will be shown along side each datapoint.
            datapoints (list[list[str]]): Outher list is the datapoints, inner list is the options for the comparison - each datapoint will be labeled.
            data_type (str, optional): The data type of the datapoints. Defaults to "media" (any form of image, video or audio). \n
                Other option: "text".
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the comparison. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be matched up with the datapoints using the list index.
            media_contexts (list[str], optional): The list of media contexts i.e. links to the images / videos for the comparison. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be matched up with the datapoints using the list index.
            a_b_names (list[str], optional): Custom naming for the two opposing models defined by the index in the datapoints list. Defaults to None.\n
                If provided has to be a list of exactly two strings.
                example:
                ```python
                datapoints = [["path_to_image_A", "path_to_image_B"], ["path_to_text_A", "path_to_text_B"]]
                a_b_naming = ["Model A", "Model B"]
                ```
                The results will then correctly show "Model A" and "Model B".
                If not provided, the results will be shown as "A" and "B".
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            confidence_threshold (float, optional): The probability threshold for the comparison. Defaults to None.\n
                If provided, the comparison datapoint will stop after the threshold is reached or at the number of responses, whatever happens first.
            filters (Sequence[RapidataFilter], optional): The list of filters for the comparison. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the comparison. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the comparison. Defaults to []. Decides in what order the tasks should be shown.
            private_notes (list[str], optional): The list of private notes for the comparison. Defaults to None.\n
                If provided has to be the same length as datapoints.\n
                This will NOT be shown to the labelers but will be included in the result purely for your own reference.
        """
        with tracer.start_as_current_span("RapidataOrderManager.create_compare_order"):
            if any(not isinstance(datapoint, list) for datapoint in datapoints):
                raise ValueError("Each datapoint must be a list of 2 paths/texts")

            if any(len(set(datapoint)) != 2 for datapoint in datapoints):
                raise ValueError(
                    "Each datapoint must contain exactly two unique options"
                )

            if a_b_names is not None and len(a_b_names) != 2:
                raise ValueError(
                    "A_B_naming must be a list of exactly two strings or None"
                )

            datapoints_instances = DatapointsValidator.map_datapoints(
                datapoints=datapoints,
                contexts=contexts,
                media_contexts=media_contexts,
                private_notes=private_notes,
                data_type=data_type,
            )
            return self._create_general_order(
                name=name,
                workflow=CompareWorkflow(instruction=instruction, a_b_names=a_b_names),
                datapoints=datapoints_instances,
                responses_per_datapoint=responses_per_datapoint,
                validation_set_id=validation_set_id,
                confidence_threshold=confidence_threshold,
                filters=filters,
                selections=selections,
                settings=settings,
            )

    def create_ranking_order(
        self,
        name: str,
        instruction: str,
        datapoints: list[list[str]],
        comparison_budget_per_ranking: int,
        responses_per_comparison: int = 1,
        data_type: Literal["media", "text"] = "media",
        random_comparisons_ratio: float = 0.5,
        contexts: Optional[list[str]] = None,
        media_contexts: Optional[list[str]] = None,
        validation_set_id: Optional[str] = None,
        filters: Sequence[RapidataFilter] | None = None,
        settings: Sequence[RapidataSetting] | None = None,
        selections: Sequence[RapidataSelection] | None = None,
    ) -> RapidataOrder:
        """
        Create a ranking order.

        With this order you can have a multiple lists of datapoints (image, text, video, audio) be ranked based on the instruction.
        Each list will be ranked independently, based on comparison matchups.

        Args:
            name (str): The name of the order.
            instruction (str): The instruction for the ranking. Will be shown with each matchup.
            datapoints (list[list[str]]): The outer list is determines the independent rankings, the inner list is the datapoints for each ranking.
            comparison_budget_per_ranking (int): The number of comparisons that will be collected per ranking (outer list of datapoints).
            responses_per_comparison (int, optional): The number of responses that will be collected per comparison. Defaults to 1.
            data_type (str, optional): The data type of the datapoints. Defaults to "media" (any form of image, video or audio). \n
                Other option: "text".
            random_comparisons_ratio (float, optional): The ratio of random comparisons to the total number of comparisons. Defaults to 0.5.
            contexts (list[str], optional): The list of contexts for the ranking. Defaults to None.\n
                If provided has to be the same length as the outer list of datapoints and will be shown in addition to the instruction. (Therefore will be different for each ranking)
                Will be matched up with the datapoints using the list index.
            media_contexts (list[str], optional): The list of media contexts for the ranking i.e links to the images / videos. Defaults to None.\n
                If provided has to be the same length as the outer list of datapoints and will be shown in addition to the instruction. (Therefore will be different for each ranking)
                Will be matched up with the datapoints using the list index.
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            filters (Sequence[RapidataFilter], optional): The list of filters for the ranking. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the ranking. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the ranking. Defaults to []. Decides in what order the tasks should be shown.
        """
        with tracer.start_as_current_span("RapidataOrderManager.create_ranking_order"):
            if contexts and len(contexts) != len(datapoints):
                raise ValueError(
                    "Number of contexts must match the number of sets that will be ranked"
                )
            if media_contexts and len(media_contexts) != len(datapoints):
                raise ValueError(
                    "Number of media contexts must match the number of sets that will be ranked"
                )
            if not isinstance(datapoints, list) or not all(
                isinstance(dp, list) for dp in datapoints
            ):
                raise ValueError(
                    "Datapoints must be a list of lists. Outer list is the independent rankings, inner list is the datapoints for each ranking."
                )
            if not all(len(set(dp)) == len(dp) for dp in datapoints):
                raise ValueError("Each inner list must contain unique datapoints.")

            if not all(len(inner_list) >= 2 for inner_list in datapoints):
                raise ValueError(
                    "Each ranking must contain at least two unique datapoints."
                )

            datapoints_instances = []
            for i, datapoint in enumerate(datapoints):
                for d in datapoint:
                    datapoints_instances.append(
                        Datapoint(
                            asset=d,
                            data_type=data_type,
                            context=contexts[i] if contexts else None,
                            media_context=media_contexts[i] if media_contexts else None,
                            group=str(i),
                        )
                    )

            contexts_dict = (
                {str(i): context for i, context in enumerate(contexts)}
                if contexts
                else None
            )

            media_contexts_dict = (
                {
                    str(i): MultiAssetInputAssetsInner(
                        actual_instance=ExistingAssetInput(
                            _t="ExistingAssetInput",
                            name=self._asset_uploader.upload_asset(media_context),
                        ),
                    )
                    for i, media_context in enumerate(media_contexts)
                }
                if media_contexts
                else None
            )

            return self._create_general_order(
                name=name,
                workflow=MultiRankingWorkflow(
                    instruction=instruction,
                    comparison_budget_per_ranking=comparison_budget_per_ranking,
                    random_comparisons_ratio=random_comparisons_ratio,
                    contexts=contexts_dict,
                    media_contexts=media_contexts_dict,
                ),
                datapoints=datapoints_instances,
                responses_per_datapoint=responses_per_comparison,
                validation_set_id=validation_set_id,
                filters=filters,
                selections=selections,
                settings=settings,
            )

    def create_free_text_order(
        self,
        name: str,
        instruction: str,
        datapoints: list[str],
        data_type: Literal["media", "text"] = "media",
        responses_per_datapoint: int = 10,
        contexts: list[str] | None = None,
        media_contexts: list[str] | None = None,
        filters: Sequence[RapidataFilter] | None = None,
        settings: Sequence[RapidataSetting] | None = None,
        selections: Sequence[RapidataSelection] | None = None,
        private_notes: list[str] | None = None,
    ) -> RapidataOrder:
        """Create a free text order.

        With this order you can have a datapoint (image, text, video, audio) be labeled with free text.
        The annotators will be shown a datapoint and will be asked to answer a question with free text.

        Args:
            name (str): The name of the order.
            instruction (str): The instruction to answer with free text. Will be shown along side each datapoint.
            datapoints (list[str]): The list of datapoints for the free text - each datapoint will be labeled.
            data_type (str, optional): The data type of the datapoints. Defaults to "media" (any form of image, video or audio). \n
                Other option: "text".
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the free text. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be matched up with the datapoints using the list index.
            media_contexts (list[str], optional): The list of media contexts for the free text i.e links to the images / videos. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be matched up with the datapoints using the list index.
            filters (Sequence[RapidataFilter], optional): The list of filters for the free text. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the free text. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the free text. Defaults to []. Decides in what order the tasks should be shown.
            private_notes (list[str], optional): The list of private notes for the free text. Defaults to None.\n
                If provided has to be the same length as datapoints.\n
                This will NOT be shown to the labelers but will be included in the result purely for your own reference.
        """
        with tracer.start_as_current_span(
            "RapidataOrderManager.create_free_text_order"
        ):
            datapoints_instances = DatapointsValidator.map_datapoints(
                datapoints=datapoints,
                contexts=contexts,
                media_contexts=media_contexts,
                private_notes=private_notes,
                data_type=data_type,
            )
            return self._create_general_order(
                name=name,
                workflow=FreeTextWorkflow(instruction=instruction),
                datapoints=datapoints_instances,
                responses_per_datapoint=responses_per_datapoint,
                filters=filters,
                selections=selections,
                settings=settings,
            )

    def create_select_words_order(
        self,
        name: str,
        instruction: str,
        datapoints: list[str],
        sentences: list[str],
        responses_per_datapoint: int = 10,
        validation_set_id: str | None = None,
        filters: Sequence[RapidataFilter] | None = None,
        settings: Sequence[RapidataSetting] | None = None,
        selections: Sequence[RapidataSelection] | None = None,
        private_notes: list[str] | None = None,
    ) -> RapidataOrder:
        """Create a select words order.

        With this order you can have a datapoint (image, text, video, audio) be labeled with a list of words.
        The annotators will be shown a datapoint as well as a list of sentences split up by spaces.
        They will then select specific words based on the instruction.

        Args:
            name (str): The name of the order.
            instruction (str): The instruction for how the words should be selected. Will be shown along side each datapoint.
            datapoints (list[str]): The list of datapoints for the select words - each datapoint will be labeled.
            sentences (list[str]): The list of sentences for the select words - Will be split up by spaces and shown along side each datapoint.\n
                Must be the same length as datapoints.
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            filters (Sequence[RapidataFilter], optional): The list of filters for the select words. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the select words. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the select words. Defaults to []. Decides in what order the tasks should be shown.
            private_notes (list[str], optional): The list of private notes for the select words. Defaults to None.\n
                If provided has to be the same length as datapoints.\n
                This will NOT be shown to the labelers but will be included in the result purely for your own reference.
        """
        with tracer.start_as_current_span(
            "RapidataOrderManager.create_select_words_order"
        ):
            datapoints_instances = DatapointsValidator.map_datapoints(
                datapoints=datapoints,
                sentences=sentences,
                private_notes=private_notes,
            )
            return self._create_general_order(
                name=name,
                workflow=SelectWordsWorkflow(
                    instruction=instruction,
                ),
                datapoints=datapoints_instances,
                responses_per_datapoint=responses_per_datapoint,
                validation_set_id=validation_set_id,
                filters=filters,
                selections=selections,
                settings=settings,
            )

    def create_locate_order(
        self,
        name: str,
        instruction: str,
        datapoints: list[str],
        responses_per_datapoint: int = 10,
        contexts: list[str] | None = None,
        media_contexts: list[str] | None = None,
        validation_set_id: str | None = None,
        filters: Sequence[RapidataFilter] | None = None,
        settings: Sequence[RapidataSetting] | None = None,
        selections: Sequence[RapidataSelection] | None = None,
        private_notes: list[str] | None = None,
    ) -> RapidataOrder:
        """Create a locate order.

        With this order you can have people locate specific objects in a datapoint (image, text, video, audio).
        The annotators will be shown a datapoint and will be asked to select locations based on the instruction.

        Args:
            name (str): The name of the order.
            instruction (str): The instruction what should be located. Will be shown along side each datapoint.
            datapoints (list[str]): The list of datapoints for the locate - each datapoint will be labeled.
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the locate. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            media_contexts (list[str], optional): The list of media contexts for the locate i.e links to the images / videos. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            filters (Sequence[RapidataFilter], optional): The list of filters for the locate. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the locate. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the locate. Defaults to []. Decides in what order the tasks should be shown.
            private_notes (list[str], optional): The list of private notes for the locate. Defaults to None.\n
                If provided has to be the same length as datapoints.\n
                This will NOT be shown to the labelers but will be included in the result purely for your own reference.
        """
        with tracer.start_as_current_span("RapidataOrderManager.create_locate_order"):
            datapoints_instances = DatapointsValidator.map_datapoints(
                datapoints=datapoints,
                contexts=contexts,
                media_contexts=media_contexts,
                private_notes=private_notes,
            )
            return self._create_general_order(
                name=name,
                workflow=LocateWorkflow(target=instruction),
                datapoints=datapoints_instances,
                responses_per_datapoint=responses_per_datapoint,
                validation_set_id=validation_set_id,
                filters=filters,
                selections=selections,
                settings=settings,
            )

    def create_draw_order(
        self,
        name: str,
        instruction: str,
        datapoints: list[str],
        responses_per_datapoint: int = 10,
        contexts: list[str] | None = None,
        media_contexts: list[str] | None = None,
        validation_set_id: str | None = None,
        filters: Sequence[RapidataFilter] = [],
        settings: Sequence[RapidataSetting] = [],
        selections: Sequence[RapidataSelection] = [],
        private_notes: list[str] | None = None,
    ) -> RapidataOrder:
        """Create a draw order.

        With this order you can have people draw lines on a datapoint (image, text, video, audio).
        The annotators will be shown a datapoint and will be asked to draw lines based on the instruction.

        Args:
            name (str): The name of the order.
            instruction (str): The instruction for how the lines should be drawn. Will be shown along side each datapoint.
            datapoints (list[str]): The list of datapoints for the draw lines - each datapoint will be labeled.
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the comparison. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            media_contexts (list[str], optional): The list of media contexts for the draw lines i.e links to the images / videos. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            filters (Sequence[RapidataFilter], optional): The list of filters for the draw lines. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the draw lines. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the draw lines. Defaults to []. Decides in what order the tasks should be shown.
            private_notes (list[str], optional): The list of private notes for the draw lines. Defaults to None.\n
                If provided has to be the same length as datapoints.\n
                This will NOT be shown to the labelers but will be included in the result purely for your own reference.
        """
        with tracer.start_as_current_span("RapidataOrderManager.create_draw_order"):
            datapoints_instances = DatapointsValidator.map_datapoints(
                datapoints=datapoints,
                contexts=contexts,
                media_contexts=media_contexts,
                private_notes=private_notes,
            )
            return self._create_general_order(
                name=name,
                workflow=DrawWorkflow(target=instruction),
                datapoints=datapoints_instances,
                responses_per_datapoint=responses_per_datapoint,
                validation_set_id=validation_set_id,
                filters=filters,
                selections=selections,
                settings=settings,
            )

    def create_timestamp_order(
        self,
        name: str,
        instruction: str,
        datapoints: list[str],
        responses_per_datapoint: int = 10,
        contexts: list[str] | None = None,
        media_contexts: list[str] | None = None,
        validation_set_id: str | None = None,
        filters: Sequence[RapidataFilter] | None = None,
        settings: Sequence[RapidataSetting] | None = None,
        selections: Sequence[RapidataSelection] | None = None,
        private_notes: list[str] | None = None,
    ) -> RapidataOrder:
        """Create a timestamp order.

        Warning:
            This order is currently not fully supported and may give unexpected results.

        With this order you can have people mark specific timestamps in a datapoint (video, audio).
        The annotators will be shown a datapoint and will be asked to select a timestamp based on the instruction.

        Args:
            name (str): The name of the order.
            instruction (str): The instruction for the timestamp task. Will be shown along side each datapoint.
            datapoints (list[str]): The list of datapoints for the timestamp - each datapoint will be labeled.
            responses_per_datapoint (int, optional): The number of responses that will be collected per datapoint. Defaults to 10.
            contexts (list[str], optional): The list of contexts for the comparison. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
                Will be match up with the datapoints using the list index.
            media_contexts (list[str], optional): The list of media contexts for the timestamp i.e links to the images / videos. Defaults to None.\n
                If provided has to be the same length as datapoints and will be shown in addition to the instruction. (Therefore will be different for each datapoint)
            validation_set_id (str, optional): The ID of the validation set. Defaults to None.\n
                If provided, one validation task will be shown infront of the datapoints that will be labeled.
            filters (Sequence[RapidataFilter], optional): The list of filters for the timestamp. Defaults to []. Decides who the tasks should be shown to.
            settings (Sequence[RapidataSetting], optional): The list of settings for the timestamp. Defaults to []. Decides how the tasks should be shown.
            selections (Sequence[RapidataSelection], optional): The list of selections for the timestamp. Defaults to []. Decides in what order the tasks should be shown.
            private_notes (list[str], optional): The list of private notes for the timestamp. Defaults to None.\n
                If provided has to be the same length as datapoints.\n
                This will NOT be shown to the labelers but will be included in the result purely for your own reference.
        """

        with tracer.start_as_current_span(
            "RapidataOrderManager.create_timestamp_order"
        ):
            datapoints_instances = DatapointsValidator.map_datapoints(
                datapoints=datapoints,
                contexts=contexts,
                media_contexts=media_contexts,
                private_notes=private_notes,
            )
            return self._create_general_order(
                name=name,
                workflow=TimestampWorkflow(instruction=instruction),
                datapoints=datapoints_instances,
                responses_per_datapoint=responses_per_datapoint,
                validation_set_id=validation_set_id,
                filters=filters,
                selections=selections,
                settings=settings,
            )

    def get_order_by_id(self, order_id: str) -> RapidataOrder:
        """Get an order by ID.

        Args:
            order_id (str): The ID of the order.

        Returns:
            RapidataOrder: The Order instance.
        """
        with tracer.start_as_current_span("RapidataOrderManager.get_order_by_id"):
            order = self.__openapi_service.order_api.order_order_id_get(order_id)

            return RapidataOrder(
                order_id=order_id,
                name=order.order_name,
                openapi_service=self.__openapi_service,
            )

    def find_orders(self, name: str = "", amount: int = 10) -> list[RapidataOrder]:
        """Find your recent orders given criteria. If nothing is provided, it will return the most recent order.

        Args:
            name (str, optional): The name of the order - matching order will contain the name. Defaults to "" for any order.
            amount (int, optional): The amount of orders to return. Defaults to 10.

        Returns:
            list[RapidataOrder]: A list of RapidataOrder instances.
        """
        with tracer.start_as_current_span("RapidataOrderManager.find_orders"):
            order_page_result = self.__openapi_service.order_api.orders_get(
                QueryModel(
                    page=PageInfo(index=1, size=amount),
                    filter=RootFilter(
                        filters=[
                            Filter(
                                field="OrderName",
                                operator=FilterOperator.CONTAINS,
                                value=name,
                            )
                        ]
                    ),
                    sortCriteria=[
                        SortCriterion(
                            direction=SortDirection.DESC, propertyName="OrderDate"
                        )
                    ],
                )
            )

            orders = [
                self.get_order_by_id(order.id) for order in order_page_result.items
            ]
            return orders

    def __str__(self) -> str:
        return "RapidataOrderManager"

    def __repr__(self) -> str:
        return self.__str__()



from rapidata.rapidata_client.settings.models.translation_behaviour_options import (
    TranslationBehaviourOptions,
)
from rapidata.rapidata_client.settings._rapidata_setting import RapidataSetting


class TranslationBehaviour(RapidataSetting):
    """
    Defines what's the behaviour of the translation in the UI.
    Will not translate text datapoints or sentences.

    Args:
        value (TranslationBehaviourOptions): The translation behaviour.
    """

    def __init__(self, value: TranslationBehaviourOptions):
        if not isinstance(value, TranslationBehaviourOptions):
            raise ValueError("The value must be a TranslationBehaviourOptions.")

        super().__init__(key="translation_behaviour", value=value.value)



from rapidata.rapidata_client.config.logger import logger
from rapidata.rapidata_client.selection._base_selection import RapidataSelection
from rapidata.api_client.models.conditional_validation_selection_validation_chance import (
    ConditionalValidationSelectionValidationChance,
)
from rapidata.api_client.models.conditional_validation_selection import (
    ConditionalValidationSelection as ConditionalValidationSelectionModel,
)
from typing import Optional


class ConditionalValidationSelection(RapidataSelection):
    """Conditional validation selection class.

    Probabilistically decides how many validation rapids you want to show per session based on the user score.

    Args:
        validation_set_id (str): The id of the validation set to be used.
        thresholds (list[float]): The thresholds to use for the user score.
        chances (list[float]): The chances of showing a validation rapid for each threshold.
        rapid_counts (list[int]): The amount of validation rapids that will be shown per session of this validation set for each threshold if selected by probability. (all or nothing)
        dimensions (Optional[list[str]], optional): The dimensions of the userScore that will be used in the thresholds. Defaults to None.

    Example:
        ```python
        ConditionalValidationSelection(
            validation_set_id="validation_set_id",
            thresholds=[0, 0.7], # (0 must be the first threshold)
            chances=[1, 0.2],
            rapid_counts=[1, 1]
        )
        ```
        This means that there's a 100% chance of showing a validation rapid if the user score is between 0 and 0.7,
        and a 20% chance of showing a validation rapid if the user score is between 0.7 and 1.
    """

    def __init__(
        self,
        validation_set_id: str,
        thresholds: list[float],
        chances: list[float],
        rapid_counts: list[int],
        dimension: Optional[str] = None,
        dimensions: Optional[list[str]] = None,
    ):
        if len(thresholds) != len(chances) or len(thresholds) != len(rapid_counts):
            raise ValueError(
                "The lengths of thresholds, chances and rapid_counts must be equal."
            )

        if dimension:
            logger.warning("dimension is deprecated, use dimensions instead")
            dimensions = (dimensions or []) + [dimension]

        self.validation_set_id = validation_set_id
        self.thresholds = thresholds
        self.chances = chances
        self.rapid_counts = rapid_counts
        self.dimensions = dimensions

    def _to_model(self):
        return ConditionalValidationSelectionModel(
            _t="ConditionalValidationSelection",
            validationSetId=self.validation_set_id,
            validationChances=[
                ConditionalValidationSelectionValidationChance(
                    userScoreThreshold=threshold, chance=chance, rapidCount=rapid_count
                )
                for threshold, chance, rapid_count in zip(
                    self.thresholds, self.chances, self.rapid_counts
                )
            ],
            dimensions=self.dimensions,
        )

    def __str__(self) -> str:
        return f"ConditionalValidationSelection(validation_set_id={self.validation_set_id}, thresholds={self.thresholds}, chances={self.chances}, rapid_counts={self.rapid_counts}, dimensions={self.dimensions})"

    def __repr__(self) -> str:
        return self.__str__()



from abc import abstractmethod
from typing import Any


class RapidataSelection:

    @abstractmethod
    def _to_model(self) -> Any:
        pass

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"



from typing import Any
from rapidata.api_client.models.attach_category_rapid_blueprint import (
    AttachCategoryRapidBlueprint,
)
from rapidata.api_client.models.simple_workflow_model import SimpleWorkflowModel
from rapidata.api_client.models.attach_category_rapid_blueprint_category import (
    AttachCategoryRapidBlueprintCategory,
)
from rapidata.api_client.models.simple_workflow_model_blueprint import (
    SimpleWorkflowModelBlueprint,
)
from rapidata.rapidata_client.workflow import Workflow
from rapidata.api_client import ClassifyPayload
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.api_client.models.rapid_modality import RapidModality
from rapidata.api_client.models.classify_payload_category import ClassifyPayloadCategory


class ClassifyWorkflow(Workflow):
    """
    A workflow for classification tasks.

    This class represents a classification workflow where a question is presented
    along with a list of possible options for classification.

    Args:
        question (str): The classification question to be presented.
        options (list[str]): A list of possible classification options.

    Attributes:
        _question (str): The classification question.
        _options (list[str]): The list of classification options.
    """

    modality = RapidModality.CLASSIFY

    def __init__(self, instruction: str, answer_options: list[str]):
        super().__init__(type="SimpleWorkflowConfig")
        self._instruction = instruction
        self._answer_options = answer_options

    def _get_instruction(self) -> str:
        return self._instruction

    def _to_dict(self) -> dict[str, Any]:
        return {
            **super()._to_dict(),
            "blueprint": {
                "_t": "ClassifyBlueprint",
                "title": self._instruction,
                "possibleCategories": self._answer_options,
            },
        }

    def _to_model(self) -> SimpleWorkflowModel:
        blueprint = AttachCategoryRapidBlueprint(
            _t="ClassifyBlueprint",
            title=self._instruction,
            categories=[
                AttachCategoryRapidBlueprintCategory(label=option, value=option)
                for option in self._answer_options
            ],
        )

        return SimpleWorkflowModel(
            _t="SimpleWorkflow",
            blueprint=SimpleWorkflowModelBlueprint(blueprint),
        )

    def _to_payload(self, datapoint: Datapoint) -> ClassifyPayload:
        return ClassifyPayload(
            _t="ClassifyPayload",
            categories=[
                ClassifyPayloadCategory(label=option, value=option)
                for option in self._answer_options
            ],
            title=self._instruction,
        )

    def __str__(self) -> str:
        return f"ClassifyWorkflow(instruction='{self._instruction}', options={self._answer_options})"

    def __repr__(self) -> str:
        return f"ClassifyWorkflow(instruction='{self._instruction}', answer_options={self._answer_options!r})"



from rapidata.api_client.models.evaluation_workflow_model import EvaluationWorkflowModel
from rapidata.rapidata_client.workflow._base_workflow import Workflow
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.api_client.models.rapid_modality import RapidModality


class EvaluationWorkflow(Workflow):
    """
    A workflow to run evaluation orders.

    This is used internally only and should not be necessary to be used by clients.

    Args:
        validation_set_id (str): a source for the tasks that will be sent to the user
        should_accept_incorrect (bool): indicates if the user should get feedback on their answers if they answer wrong. If set to true the user will not notice that he was tested.
    """

    modality = RapidModality.NONE

    def __init__(self, validation_set_id: str, should_accept_incorrect: bool):
        super().__init__("EvaluationWorkflow")
        self.validation_set_id = validation_set_id
        self.should_accept_incorrect = should_accept_incorrect

    def _get_instruction(self) -> str:
        return ""

    def _to_model(self):
        return EvaluationWorkflowModel(
            _t="EvaluationWorkflow",
            validationSetId=self.validation_set_id,
            shouldAcceptIncorrect=self.should_accept_incorrect,
        )

    def _to_payload(self, datapoint: Datapoint):
        raise NotImplementedError("EvaluationWorkflow does not have a payload")

    def __str__(self) -> str:
        return f"EvaluationWorkflow(validation_set_id='{self.validation_set_id}', should_accept_incorrect={self.should_accept_incorrect})"

    def __repr__(self) -> str:
        return f"EvaluationWorkflow(validation_set_id={self.validation_set_id!r}, should_accept_incorrect={self.should_accept_incorrect!r})"



# coding: utf-8

"""
    Rapidata.Dataset

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: v1
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from rapidata.api_client.models.get_validation_rapids_result_asset import GetValidationRapidsResultAsset
from rapidata.api_client.models.get_validation_rapids_result_payload import GetValidationRapidsResultPayload
from rapidata.api_client.models.get_validation_rapids_result_truth import GetValidationRapidsResultTruth
from rapidata.api_client.models.rapid_state import RapidState
from typing import Optional, Set
from typing_extensions import Self

class GetValidationRapidsResult(BaseModel):
    """
    GetValidationRapidsResult
    """ # noqa: E501
    id: StrictStr
    type: StrictStr
    asset: Optional[GetValidationRapidsResultAsset] = None
    truth: Optional[GetValidationRapidsResultTruth] = None
    payload: GetValidationRapidsResultPayload
    context: Optional[StrictStr] = None
    context_asset: Optional[GetValidationRapidsResultAsset] = Field(default=None, alias="contextAsset")
    correct_validation_count: StrictInt = Field(alias="correctValidationCount")
    invalid_validation_count: StrictInt = Field(alias="invalidValidationCount")
    explanation: Optional[StrictStr] = None
    random_correct_probability: Optional[Union[StrictFloat, StrictInt]] = Field(default=None, alias="randomCorrectProbability")
    state: RapidState
    __properties: ClassVar[List[str]] = ["id", "type", "asset", "truth", "payload", "context", "contextAsset", "correctValidationCount", "invalidValidationCount", "explanation", "randomCorrectProbability", "state"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of GetValidationRapidsResult from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of asset
        if self.asset:
            _dict['asset'] = self.asset.to_dict()
        # override the default output from pydantic by calling `to_dict()` of truth
        if self.truth:
            _dict['truth'] = self.truth.to_dict()
        # override the default output from pydantic by calling `to_dict()` of payload
        if self.payload:
            _dict['payload'] = self.payload.to_dict()
        # override the default output from pydantic by calling `to_dict()` of context_asset
        if self.context_asset:
            _dict['contextAsset'] = self.context_asset.to_dict()
        # set to None if asset (nullable) is None
        # and model_fields_set contains the field
        if self.asset is None and "asset" in self.model_fields_set:
            _dict['asset'] = None

        # set to None if truth (nullable) is None
        # and model_fields_set contains the field
        if self.truth is None and "truth" in self.model_fields_set:
            _dict['truth'] = None

        # set to None if context (nullable) is None
        # and model_fields_set contains the field
        if self.context is None and "context" in self.model_fields_set:
            _dict['context'] = None

        # set to None if context_asset (nullable) is None
        # and model_fields_set contains the field
        if self.context_asset is None and "context_asset" in self.model_fields_set:
            _dict['contextAsset'] = None

        # set to None if explanation (nullable) is None
        # and model_fields_set contains the field
        if self.explanation is None and "explanation" in self.model_fields_set:
            _dict['explanation'] = None

        # set to None if random_correct_probability (nullable) is None
        # and model_fields_set contains the field
        if self.random_correct_probability is None and "random_correct_probability" in self.model_fields_set:
            _dict['randomCorrectProbability'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of GetValidationRapidsResult from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id": obj.get("id"),
            "type": obj.get("type"),
            "asset": GetValidationRapidsResultAsset.from_dict(obj["asset"]) if obj.get("asset") is not None else None,
            "truth": GetValidationRapidsResultTruth.from_dict(obj["truth"]) if obj.get("truth") is not None else None,
            "payload": GetValidationRapidsResultPayload.from_dict(obj["payload"]) if obj.get("payload") is not None else None,
            "context": obj.get("context"),
            "contextAsset": GetValidationRapidsResultAsset.from_dict(obj["contextAsset"]) if obj.get("contextAsset") is not None else None,
            "correctValidationCount": obj.get("correctValidationCount"),
            "invalidValidationCount": obj.get("invalidValidationCount"),
            "explanation": obj.get("explanation"),
            "randomCorrectProbability": obj.get("randomCorrectProbability"),
            "state": obj.get("state")
        })
        return _obj





import webbrowser
import urllib.parse
from colorama import Fore
from rapidata.rapidata_client.validation.rapids.rapids import Rapid
from rapidata.service.openapi_service import OpenAPIService
from rapidata.rapidata_client.config import logger, managed_print, tracer
from rapidata.api_client.models.update_validation_set_model import (
    UpdateValidationSetModel,
)
from rapidata.api_client.models.update_should_alert_model import UpdateShouldAlertModel
from rapidata.rapidata_client.validation.rapids._validation_rapid_uploader import (
    ValidationRapidUploader,
)


class RapidataValidationSet:
    """A class for interacting with a Rapidata validation set.

    Represents a set of all the validation tasks that can be added to an order.

    When added to an order, the tasks will be selected randomly from the set.

    Attributes:
        id (str): The ID of the validation set.
        name (str): The name of the validation set.
    """

    def __init__(
        self,
        validation_set_id,
        name: str,
        dimensions: list[str],
        openapi_service: OpenAPIService,
    ):
        self.id = validation_set_id
        self.name = name
        self.dimensions = dimensions
        self.validation_set_details_page = (
            f"https://app.{openapi_service.environment}/validation-set/detail/{self.id}"
        )
        self._openapi_service = openapi_service
        self.validation_rapid_uploader = ValidationRapidUploader(openapi_service)

    def add_rapid(self, rapid: Rapid):
        """Add a Rapid to the validation set.

        Args:
            rapid (Rapid): The Rapid to add to the validation set.
        """
        with tracer.start_as_current_span("RapidataValidationSet.add_rapid"):
            logger.debug("Adding rapid %s to validation set %s", rapid, self.id)
            self.validation_rapid_uploader.upload_rapid(rapid, self.id)
        return self

    def update_dimensions(self, dimensions: list[str]):
        """Update the dimensions of the validation set.

        Args:
            dimensions (list[str]): The new dimensions of the validation set.
        """
        with tracer.start_as_current_span("RapidataValidationSet.update_dimensions"):
            logger.debug(
                "Updating dimensions for validation set %s to %s", self.id, dimensions
            )
            self._openapi_service.validation_api.validation_set_validation_set_id_patch(
                self.id, UpdateValidationSetModel(dimensions=dimensions)
            )
            self.dimensions = dimensions
            return self

    def update_should_alert(self, should_alert: bool):
        """Determines whether users should be alerted if they answer incorrectly.

        Args:
            should_alert (bool): Specifies whether users should be alerted for incorrect answers. Defaults to True if not specifically overridden by this method.

        Note:
            The userScore dimensions which are updated when a user answers a validation task are updated regardless of the value of `should_alert`.
        """
        with tracer.start_as_current_span("RapidataValidationSet.update_should_alert"):
            logger.debug(
                "Setting shouldAlert for validation set %s to %s", self.id, should_alert
            )
            self._openapi_service.validation_api.validation_set_validation_set_id_patch(
                self.id, UpdateValidationSetModel(shouldAlert=should_alert)
            )
            return self

    def update_can_be_flagged(self, can_be_flagged: bool):
        """Update if tasks in the validation set can be flagged for bad accuracy.

        Args:
            can_be_flagged (bool): Specifies whether tasks in the validation set can be flagged for bad accuracy. Defaults to True if not specifically overridden by this method.
        """
        with tracer.start_as_current_span(
            "RapidataValidationSet.update_can_be_flagged"
        ):
            logger.debug(
                "Setting canBeFlagged for validation set %s to %s",
                self.id,
                can_be_flagged,
            )
            self._openapi_service.validation_api.validation_set_validation_set_id_patch(
                self.id, UpdateValidationSetModel(isFlagOverruled=(not can_be_flagged))
            )
            return self

    def view(self) -> None:
        """
        Opens the validation set details page in the browser.

        Raises:
            Exception: If the order is not in processing state.
        """
        logger.info("Opening validation set details page in browser...")
        could_open_browser = webbrowser.open(self.validation_set_details_page)
        if not could_open_browser:
            encoded_url = urllib.parse.quote(
                self.validation_set_details_page, safe="%/:=&?~#+!$,;'@()*[]"
            )
            managed_print(
                Fore.RED
                + f"Please open this URL in your browser: '{encoded_url}'"
                + Fore.RESET
            )

    def delete(self) -> None:
        """Deletes the validation set"""
        with tracer.start_as_current_span("RapidataValidationSet.delete"):
            logger.info("Deleting ValidationSet '%s'", self)
            self._openapi_service.validation_api.validation_set_validation_set_id_delete(
                self.id
            )
            logger.debug("ValidationSet '%s' has been deleted.", self)
            managed_print(f"ValidationSet '{self}' has been deleted.")

    def __str__(self):
        return f"name: '{self.name}' id: {self.id} dimensions: {self.dimensions}"

    def __repr__(self):
        return f"name: '{self.name}' id: {self.id} dimensions: {self.dimensions}"



from rapidata.rapidata_client.filter._base_filter import RapidataFilter
from rapidata.api_client.models.new_user_filter_model import NewUserFilterModel


class NewUserFilter(RapidataFilter):
    """NewUserFilter Class

    Can be used to filter new users.
    """

    def _to_model(self):
        return NewUserFilterModel(_t="NewUserFilter")



from rapidata.api_client.models.box_shape import BoxShape
from pydantic import BaseModel, field_validator, model_validator


class Box(BaseModel):
    """
    Used in the Locate and Draw Validation sets. All coordinates are in ratio of the image size (0.0 to 1.0).

    Args:
        x_min (float): The minimum x value of the box in ratio of the image size.
        y_min (float): The minimum y value of the box in ratio of the image size.
        x_max (float): The maximum x value of the box in ratio of the image size.
        y_max (float): The maximum y value of the box in ratio of the image size.
    """

    x_min: float
    y_min: float
    x_max: float
    y_max: float

    @field_validator("x_min", "y_min", "x_max", "y_max")
    @classmethod
    def coordinates_between_zero_and_one(cls, v: float) -> float:
        if not (0.0 <= v <= 1.0):
            raise ValueError("Box coordinates must be between 0 and 1")
        return v

    @model_validator(mode="after")
    def check_min_less_than_max(self) -> "Box":
        if self.x_min >= self.x_max:
            raise ValueError("x_min must be less than x_max")
        if self.y_min >= self.y_max:
            raise ValueError("y_min must be less than y_max")
        return self

    def to_model(self) -> BoxShape:
        return BoxShape(
            _t="BoxShape",
            xMin=self.x_min * 100,
            yMin=self.y_min * 100,
            xMax=self.x_max * 100,
            yMax=self.y_max * 100,
        )



from rapidata.api_client import (
    CompareWorkflowModelPairMakerConfig,
    OnlinePairMakerConfigModel,
    EloConfigModel,
)
from rapidata.api_client.models.compare_workflow_model import CompareWorkflowModel
from rapidata.rapidata_client.workflow._base_workflow import Workflow
from rapidata.api_client import ComparePayload
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.api_client.models.rapid_modality import RapidModality
from rapidata.rapidata_client.datapoints.metadata import (
    MediaAssetMetadata,
    PromptMetadata,
)
from rapidata.api_client.models.compare_workflow_model_metadata_inner import (
    CompareWorkflowModelMetadataInner,
)
from rapidata.rapidata_client.datapoints._asset_uploader import AssetUploader
import itertools
import random
from typing import cast


class RankingWorkflow(Workflow):
    modality = RapidModality.COMPARE

    def __init__(
        self,
        instruction: str,
        total_comparison_budget: int,
        random_comparisons_ratio,
        elo_start: int = 1200,
        elo_k_factor: int = 40,
        elo_scaling_factor: int = 400,
        media_context: str | None = None,
        context: str | None = None,
        file_uploader: AssetUploader | None = None,
    ):
        super().__init__(type="CompareWorkflowConfig")

        self.media_context = media_context
        self.context = context

        self.metadatas = []
        if media_context:
            assert (
                file_uploader is not None
            ), "File uploader is required if media_context is provided"
            self.metadatas.append(
                MediaAssetMetadata(
                    internal_file_name=file_uploader.upload_asset(media_context)
                )
            )
        if context:
            self.metadatas.append(PromptMetadata(prompt=context))

        self.instruction = instruction
        self.total_comparison_budget = total_comparison_budget
        self.random_comparisons_ratio = random_comparisons_ratio
        self.elo_start = elo_start
        self.elo_k_factor = elo_k_factor
        self.elo_scaling_factor = elo_scaling_factor

        self.pair_maker_config = CompareWorkflowModelPairMakerConfig(
            OnlinePairMakerConfigModel(
                _t="OnlinePairMaker",
                totalComparisonBudget=total_comparison_budget,
                randomMatchesRatio=random_comparisons_ratio,
            )
        )

        self.elo_config = EloConfigModel(
            startingElo=elo_start,
            kFactor=elo_k_factor,
            scalingFactor=elo_scaling_factor,
        )

    def _get_instruction(self) -> str:
        return self.instruction

    def _to_model(self) -> CompareWorkflowModel:

        return CompareWorkflowModel(
            _t="CompareWorkflow",
            criteria=self.instruction,
            eloConfig=self.elo_config,
            pairMakerConfig=self.pair_maker_config,
            metadata=[
                CompareWorkflowModelMetadataInner(metadata.to_model())
                for metadata in self.metadatas
            ],
        )

    def _to_payload(self, datapoint: Datapoint) -> ComparePayload:
        return ComparePayload(
            _t="ComparePayload",
            criteria=self.instruction,
        )

    def _format_datapoints(self, datapoints: list[Datapoint]) -> list[Datapoint]:
        if len(datapoints) < 3:
            raise ValueError("RankingWorkflow requires at least three datapoints")
        desired_length = len(datapoints)
        assets = [datapoint.asset for datapoint in datapoints]
        pairs = list(map(list, itertools.combinations(assets, 2)))
        sampled_pairs = random.sample(pairs, desired_length)
        formatted_datapoints = [
            Datapoint(
                asset=cast(list[str], pair),
                data_type=datapoints[0].data_type,
                context=self.context,
                media_context=self.media_context,
            )
            for pair in sampled_pairs
        ]
        return formatted_datapoints

    def __str__(self) -> str:
        return f"RankingWorkflow(instruction='{self.instruction}', metadatas={self.metadatas})"

    def __repr__(self) -> str:
        return f"RankingWorkflow(instruction={self.instruction!r}, total_comparison_budget={self.total_comparison_budget!r}, random_comparisons_ratio={self.random_comparisons_ratio!r}, elo_start={self.elo_start!r}, elo_k_factor={self.elo_k_factor!r}, elo_scaling_factor={self.elo_scaling_factor!r}, metadatas={self.metadatas!r})"



from pydantic_core.core_schema import FieldValidationInfo
from rapidata.rapidata_client.filter._base_filter import RapidataFilter
from rapidata.api_client.models.user_score_user_filter_model import (
    UserScoreUserFilterModel,
)
from pydantic import BaseModel, field_validator, model_validator


class UserScoreFilter(RapidataFilter, BaseModel):
    """UserScoreFilter Class

    Can be used to filter who to target based on their user score.

    Args:
        lower_bound (float): The lower bound of the user score.
        upper_bound (float): The upper bound of the user score.
        dimension (str): The dimension of the userScore to be considerd for the filter.

    Example:
        ```python
        UserScoreFilter(0.5, 0.9)
        ```
        This will only show the order to users that have a UserScore of >=0.5 and <=0.9
    """

    lower_bound: float
    upper_bound: float
    dimension: str | None = None

    def __init__(
        self,
        lower_bound: float = 0.0,
        upper_bound: float = 1.0,
        dimension: str | None = None,
    ):
        super().__init__(
            lower_bound=lower_bound, upper_bound=upper_bound, dimension=dimension
        )

    @field_validator("lower_bound", "upper_bound")
    @classmethod
    def validate_bounds(cls, v: float, info: FieldValidationInfo) -> float:
        if v < 0 or v > 1:
            raise ValueError(f"{info.field_name} must be between 0 and 1")
        return v

    @model_validator(mode="after")
    def validate_bounds_relationship(self) -> "UserScoreFilter":
        if self.lower_bound >= self.upper_bound:
            raise ValueError("lower_bound must be less than upper_bound")
        return self

    def _to_model(self):
        return UserScoreUserFilterModel(
            _t="UserScoreFilter",
            upperbound=self.upper_bound,
            lowerbound=self.lower_bound,
            dimension=self.dimension,
        )



from rapidata.api_client.models.ab_test_selection_a_inner import AbTestSelectionAInner
from rapidata.api_client.models.shuffling_selection import (
    ShufflingSelection as ShufflingSelectionModel,
)
from rapidata.rapidata_client.selection._base_selection import RapidataSelection

from typing import Sequence


class ShufflingSelection(RapidataSelection):
    """ShufflingSelection Class

    Shuffles the selections provided in the list.

    Args:
        selections (Sequence[RapidataSelection]): List of selections to shuffle.

    Example:
        ```python
        selection = ShufflingSelection(
                    [ValidSelections("validation_id", 1), LabelingSelection(2)])
        ```
        This means that the users will get 1 validation task and 2 labeling tasks in a shuffled order.
    """

    def __init__(self, selections: Sequence[RapidataSelection]):
        self.selections = selections

    def _to_model(self) -> ShufflingSelectionModel:
        return ShufflingSelectionModel(
            _t="ShufflingSelection",
            selections=[
                AbTestSelectionAInner(selection._to_model())
                for selection in self.selections
            ],
        )

    def __str__(self) -> str:
        return f"ShufflingSelection(selections={self.selections})"

    def __repr__(self) -> str:
        return self.__str__()



from rapidata.api_client.models.simple_workflow_model import SimpleWorkflowModel
from rapidata.api_client.models.validation_set_zip_post_request_blueprint import (
    ValidationSetZipPostRequestBlueprint,
)
from rapidata.api_client.models.simple_workflow_model_blueprint import (
    SimpleWorkflowModelBlueprint,
)
from rapidata.api_client.models.scrub_rapid_blueprint import ScrubRapidBlueprint
from rapidata.rapidata_client.workflow._base_workflow import Workflow
from rapidata.api_client import ScrubPayload
from rapidata.rapidata_client.datapoints._datapoint import Datapoint
from rapidata.api_client.models.rapid_modality import RapidModality


class TimestampWorkflow(Workflow):
    """
    A workflow for timestamp tasks.

    This class represents a timestamp workflow
    where audio or video content receives timestamps based on the instruction.

    Attributes:
        _instruction (str): The instruction for the timestamp task.

    Args:
        instruction (str): The instruction to be provided for the timestamp task.
    """

    modality = RapidModality.SCRUB

    def __init__(self, instruction: str):
        super().__init__(type="SimpleWorkflowConfig")
        self._instruction = instruction

    def _get_instruction(self) -> str:
        return self._instruction

    def _to_model(self) -> SimpleWorkflowModel:
        blueprint = ScrubRapidBlueprint(_t="ScrubBlueprint", target=self._instruction)

        return SimpleWorkflowModel(
            _t="SimpleWorkflow",
            blueprint=SimpleWorkflowModelBlueprint(blueprint),
        )

    def _to_payload(self, datapoint: Datapoint) -> ScrubPayload:
        return ScrubPayload(
            _t="ScrubPayload",
            target=self._instruction,
        )

    def __str__(self) -> str:
        return f"TimestampWorkflow(instruction='{self._instruction}')"

    def __repr__(self) -> str:
        return f"TimestampWorkflow(instruction={self._instruction!r})"



from rapidata.rapidata_client.validation.rapids.rapids import Rapid
from rapidata.service.openapi_service import OpenAPIService
from rapidata.api_client.models.add_validation_rapid_model import (
    AddValidationRapidModel,
)
from rapidata.api_client.models.add_validation_rapid_model_truth import (
    AddValidationRapidModelTruth,
)
from rapidata.rapidata_client.datapoints._asset_uploader import AssetUploader
from rapidata.api_client.models.add_validation_rapid_model_asset import (
    AddValidationRapidModelAsset,
)
from rapidata.api_client.models.add_validation_rapid_model_context_asset import (
    AddValidationRapidModelContextAsset,
)
from rapidata.api_client.models.add_validation_rapid_model_payload import (
    AddValidationRapidModelPayload,
)


class ValidationRapidUploader:
    def __init__(self, openapi_service: OpenAPIService):
        self.openapi_service = openapi_service
        self.asset_uploader = AssetUploader(openapi_service)

    def upload_rapid(self, rapid: Rapid, validation_set_id: str) -> None:
        uploaded_asset = (
            self._handle_media_rapid(rapid)
            if rapid.data_type == "media"
            else self._handle_text_rapid(rapid)
        )

        self.openapi_service.validation_api.validation_set_validation_set_id_rapid_post(
            validation_set_id=validation_set_id,
            add_validation_rapid_model=AddValidationRapidModel(
                asset=uploaded_asset,
                context=rapid.context,
                contextAsset=(
                    AddValidationRapidModelContextAsset(
                        actual_instance=self.asset_uploader.get_uploaded_asset_input(
                            rapid.media_context
                        ),
                    )
                    if rapid.media_context
                    else None
                ),
                payload=self._get_payload(rapid),
                truth=AddValidationRapidModelTruth(actual_instance=rapid.truth),
                randomCorrectProbability=rapid.random_correct_probability,
                explanation=rapid.explanation,
                featureFlags=(
                    [setting._to_feature_flag() for setting in rapid.settings]
                    if rapid.settings
                    else None
                ),
            ),
        )

    def _get_payload(self, rapid: Rapid) -> AddValidationRapidModelPayload:
        if isinstance(rapid.payload, dict):
            return AddValidationRapidModelPayload(actual_instance=rapid.payload)
        return AddValidationRapidModelPayload(actual_instance=rapid.payload.to_dict())

    def _handle_text_rapid(self, rapid: Rapid) -> AddValidationRapidModelAsset:
        return AddValidationRapidModelAsset(
            actual_instance=self.asset_uploader.get_uploaded_text_input(rapid.asset),
        )

    def _handle_media_rapid(self, rapid: Rapid) -> AddValidationRapidModelAsset:
        return AddValidationRapidModelAsset(
            actual_instance=self.asset_uploader.get_uploaded_asset_input(rapid.asset),
        )



from typing import Literal


class DetailMapper:
    @staticmethod
    def get_budget(
        level_of_detail: Literal["low", "medium", "high", "very high"]
    ) -> int:
        if level_of_detail == "low":
            return 2_000
        elif level_of_detail == "medium":
            return 4_000
        elif level_of_detail == "high":
            return 8_000
        elif level_of_detail == "very high":
            return 16_000
        else:
            raise ValueError(
                "Invalid level of detail. Must be one of: 'low', 'medium', 'high', 'very high'"
            )

    @staticmethod
    def get_level_of_detail(
        response_budget: int,
    ) -> Literal["low", "medium", "high", "very high"]:
        if response_budget < 4_000:
            return "low"
        elif response_budget < 8_000:
            return "medium"
        elif response_budget < 16_000:
            return "high"
        return "very high"



import os
from PIL import Image


class LocalFileService:

    def load_image(self, image_path: str) -> Image.Image:
        self.check_file_exists(image_path)
        return Image.open(image_path)

    def load_images(self, image_paths: list[str]) -> list[Image.Image]:
        return [self.load_image(image_path) for image_path in image_paths]

    def load_video(self, video_path: str):
        self.check_file_exists(video_path)
        return open(video_path, "rb")

    def load_videos(self, video_paths: list[str]):
        return [self.load_video(video_path) for video_path in video_paths]

    def _file_exists(self, file_path: str) -> bool:
        return os.path.exists(file_path)

    def check_file_exists(self, file_path: str):
        if not self._file_exists(file_path):
            raise FileNotFoundError(f"File {file_path} not found.")
